@@ -7,8 +7,8 @@
  * netdump dumpfiles, the facilities in netdump.c are used.  For
  * compressed dumpfiles, the facilities in this file are used.
  *
- * Copyright (C) 2004, 2005, 2006, 2007, 2008 David Anderson
- * Copyright (C) 2004, 2005, 2006, 2007, 2008 Red Hat, Inc. All rights reserved.
+ * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 David Anderson
+ * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Red Hat, Inc. All rights reserved.
  * Copyright (C) 2005  FUJITSU LIMITED
  * Copyright (C) 2005  NEC Corporation
  *
@@ -49,6 +49,9 @@ struct diskdump_data {
 	int	byte, bit;
 	char	*compressed_page;	/* copy of compressed page data */
 	char	*curbufptr;		/* ptr to uncompressed page buffer */
+	unsigned char *notes_buf;	/* copy of elf notes */
+	void	**nt_prstatus_percpu;
+	uint	num_prstatus_notes;
 
 	/* page cache */
 	struct page_cache_hdr {		/* header for each cached page */
@@ -72,6 +75,10 @@ static int get_dump_level(void);
 ulong *diskdump_flags = &diskdump_data.flags;
 
 static int __diskdump_memory_dump(FILE *);
+static void dump_vmcoreinfo(FILE *);
+static void dump_nt_prstatus_offset(FILE *);
+static char *vmcoreinfo_read_string(const char *);
+static void diskdump_get_osrelease(void);
 
 /* For split dumpfile */
 static struct diskdump_data **dd_list = NULL;
@@ -83,7 +90,42 @@ int dumpfile_is_split(void)
 	return KDUMP_SPLIT();
 }
 
-static void add_diskdump_data(char* name)
+void
+map_cpus_to_prstatus_kdump_cmprs(void)
+{
+	void **nt_ptr;
+	int online, i, j, nrcpus;
+	size_t size;
+
+	if (!(online = get_cpus_online()) || (online == kt->cpus))
+		return;
+
+	if (CRASHDEBUG(1))
+		error(INFO,
+		    "cpus: %d online: %d NT_PRSTATUS notes: %d (remapping)\n",
+			kt->cpus, online, dd->num_prstatus_notes);
+
+	size = NR_CPUS * sizeof(void *);
+
+	nt_ptr = (void **)GETBUF(size);
+	BCOPY(dd->nt_prstatus_percpu, nt_ptr, size);
+	BZERO(dd->nt_prstatus_percpu, size);
+
+	/*
+	 *  Re-populate the array with the notes mapping to online cpus
+	 */
+	nrcpus = (kt->kernel_NR_CPUS ? kt->kernel_NR_CPUS : NR_CPUS);
+
+	for (i = 0, j = 0; i < nrcpus; i++) {
+		if (in_cpu_map(ONLINE, i))
+			dd->nt_prstatus_percpu[i] = nt_ptr[j++];
+	}
+
+	FREEBUF(nt_ptr);
+}
+
+static void 
+add_diskdump_data(char* name)
 {
 #define DDL_SIZE 16
 	int i;
@@ -130,7 +172,8 @@ static void add_diskdump_data(char* name)
 			dd->sub_header_kdump->end_pfn);
 }
 
-static void clean_diskdump_data(void)
+static void 
+clean_diskdump_data(void)
 {
 	int i;
 
@@ -146,28 +189,33 @@ static void clean_diskdump_data(void)
 	dd = &diskdump_data;
 }
 
-static inline int get_bit(char *map, int byte, int bit)
+static inline int 
+get_bit(char *map, int byte, int bit)
 {
 	return map[byte] & (1<<bit);
 }
 
-static inline int page_is_ram(unsigned int nr)
+static inline int 
+page_is_ram(unsigned int nr)
 {
 	return get_bit(dd->bitmap, nr >> 3, nr & 7);
 }
 
-static inline int page_is_dumpable(unsigned int nr)
+static inline int 
+page_is_dumpable(unsigned int nr)
 {
 	return dd->dumpable_bitmap[nr>>3] & (1 << (nr & 7));
 }
 
-static inline int dump_is_partial(const struct disk_dump_header *header)
+static inline int 
+dump_is_partial(const struct disk_dump_header *header)
 {
 	return header->bitmap_blocks >=
 	    divideup(divideup(header->max_mapnr, 8), dd->block_size) * 2;
 }
 
-static int open_dump_file(char *file)
+static int 
+open_dump_file(char *file)
 {
 	int fd;
 
@@ -184,11 +232,53 @@ static int open_dump_file(char *file)
 	return TRUE;
 }
 
-static int read_dump_header(char *file)
+void 
+x86_process_elf_notes(void *note_ptr, unsigned long size_note)
+{
+	Elf32_Nhdr *note32 = NULL;
+	Elf64_Nhdr *note64 = NULL;
+	size_t tot, len = 0;
+	int num = 0;
+
+	for (tot = 0; tot < size_note; tot += len) {
+		if (machine_type("X86_64")) {
+			note64 = note_ptr + tot;
+
+			if (note64->n_type == NT_PRSTATUS) {
+				dd->nt_prstatus_percpu[num] = note64;
+				num++;
+			}
+
+			len = sizeof(Elf64_Nhdr);
+			len = roundup(len + note64->n_namesz, 4);
+			len = roundup(len + note64->n_descsz, 4);
+		} else if (machine_type("X86")) {
+			note32 = note_ptr + tot;
+
+			if (note32->n_type == NT_PRSTATUS) {
+				dd->nt_prstatus_percpu[num] = note32;
+				num++;
+			}
+
+			len = sizeof(Elf32_Nhdr);
+			len = roundup(len + note32->n_namesz, 4);
+			len = roundup(len + note32->n_descsz, 4);
+		}
+	}
+
+	if (num > 0) {
+		pc->flags2 |= ELF_NOTES;
+		dd->num_prstatus_notes = num;
+	}
+}
+
+static int 
+read_dump_header(char *file)
 {
 	struct disk_dump_header *header = NULL;
 	struct disk_dump_sub_header *sub_header = NULL;
 	struct kdump_sub_header *sub_header_kdump = NULL;
+	size_t size;
 	int bitmap_len;
 	int block_size = (int)sysconf(_SC_PAGESIZE);
 	off_t offset;
@@ -204,16 +294,22 @@ restart:
 	if ((header = realloc(header, block_size)) == NULL)
 		error(FATAL, "diskdump / compressed kdump: cannot malloc block_size buffer\n");
 
-	if (lseek(dd->dfd, 0, SEEK_SET) == failed) {
-		if (CRASHDEBUG(1))
-			error(INFO, "diskdump / compressed kdump: cannot lseek dump header\n");
-		goto err;
-	}
-
-	if (read(dd->dfd, header, block_size) < block_size) {
-		if (CRASHDEBUG(1))
-			error(INFO, "diskdump / compressed kdump: cannot read dump header\n");
-		goto err;
+	if (FLAT_FORMAT()) {
+		if (!read_flattened_format(dd->dfd, 0, header, block_size)) {
+			error(FATAL, "diskdump / compressed kdump: cannot read header\n");
+			goto err;
+		}
+	} else {
+		if (lseek(dd->dfd, 0, SEEK_SET) == failed) {
+			if (CRASHDEBUG(1))
+				error(INFO, "diskdump / compressed kdump: cannot lseek dump header\n");
+			goto err;
+		}
+		if (read(dd->dfd, header, block_size) < block_size) {
+			if (CRASHDEBUG(1))
+				error(INFO, "diskdump / compressed kdump: cannot read dump header\n");
+			goto err;
+		}
 	}
 
 	/* validate dump header */
@@ -252,6 +348,9 @@ restart:
 	else if (STRNEQ(header->utsname.machine, "arm") &&
 	    machine_type_mismatch(file, "ARM", NULL, 0))
 		goto err;
+	else if (STRNEQ(header->utsname.machine, "s390x") &&
+	    machine_type_mismatch(file, "S390X", NULL, 0))
+		goto err;
 
 	if (header->block_size != block_size) {
 		block_size = header->block_size;
@@ -264,41 +363,56 @@ restart:
 	dd->block_size  = header->block_size;
 	dd->block_shift = ffs(header->block_size) - 1;
 
-	if (sizeof(*header) + sizeof(void *) * header->nr_cpus > block_size ||
-	    header->nr_cpus <= 0) {
-		error(INFO, "%s: invalid nr_cpus value: %d\n", 
-			DISKDUMP_VALID() ? "diskdump" : "compressed kdump",
-			header->nr_cpus);
-		goto err;
-	}
+	if ((DISKDUMP_VALID() &&
+             (sizeof(*header) + sizeof(void *) * header->nr_cpus > block_size)) ||
+             header->nr_cpus <= 0) {
+                error(INFO, "%s: invalid nr_cpus value: %d\n",
+                        DISKDUMP_VALID() ? "diskdump" : "compressed kdump",
+                        header->nr_cpus);
+                        goto err;
+        }
 
 	/* read sub header */
 	offset = (off_t)block_size;
-	if (lseek(dd->dfd, offset, SEEK_SET) == failed) {
-		error(INFO, "%s: cannot lseek dump sub header\n",
-			DISKDUMP_VALID() ? "diskdump" : "compressed kdump");
-
-		goto err;
-	}
 
 	if (DISKDUMP_VALID()) {
 		if ((sub_header = malloc(block_size)) == NULL)
 			error(FATAL, "diskdump: cannot malloc sub_header buffer\n");
 
-		if (read(dd->dfd, sub_header, block_size)
-		  < block_size) {
-			error(INFO, "diskdump: cannot read dump sub header\n");
-			goto err;
+		if (FLAT_FORMAT()) {
+			if (!read_flattened_format(dd->dfd, offset, sub_header, block_size)) {
+				error(INFO, "diskdump: cannot read dump sub header\n");
+				goto err;
+			}
+		} else {
+			if (lseek(dd->dfd, offset, SEEK_SET) == failed) {
+				error(INFO, "diskdump: cannot lseek dump sub header\n");
+				goto err;
+			}
+			if (read(dd->dfd, sub_header, block_size) < block_size) {
+				error(INFO, "diskdump: cannot read dump sub header\n");
+				goto err;
+			}
 		}
 		dd->sub_header = sub_header;
 	} else if (KDUMP_CMPRS_VALID()) {
 		if ((sub_header_kdump = malloc(block_size)) == NULL)
 			error(FATAL, "compressed kdump: cannot malloc sub_header_kdump buffer\n");
 
-		if (read(dd->dfd, sub_header_kdump, block_size)
-		  < block_size) {
-			error(INFO, "compressed kdump: cannot read dump sub header\n");
-			goto err;
+		if (FLAT_FORMAT()) {
+			if (!read_flattened_format(dd->dfd, offset, sub_header_kdump, block_size)) {
+				error(INFO, "compressed kdump: cannot read dump sub header\n");
+				goto err;
+			}
+		} else {
+			if (lseek(dd->dfd, offset, SEEK_SET) == failed) {
+				error(INFO, "compressed kdump: cannot lseek dump sub header\n");
+				goto err;
+			}
+			if (read(dd->dfd, sub_header_kdump, block_size) < block_size) {
+				error(INFO, "compressed kdump: cannot read dump sub header\n");
+				goto err;
+			}
 		}
 		dd->sub_header_kdump = sub_header_kdump;
 	}
@@ -308,22 +422,30 @@ restart:
 	dd->bitmap_len = bitmap_len;
 
 	offset = (off_t)block_size * (1 + header->sub_hdr_size);
-	if (lseek(dd->dfd, offset, SEEK_SET) == failed) {
-		error(INFO, "%s: cannot lseek memory bitmap\n",
-			DISKDUMP_VALID() ? "diskdump" : "compressed kdump");
-
-		goto err;
-	}
 
 	if ((dd->bitmap = malloc(bitmap_len)) == NULL)
 		error(FATAL, "%s: cannot malloc bitmap buffer\n",
 			DISKDUMP_VALID() ? "diskdump" : "compressed kdump");
 
 	dd->dumpable_bitmap = calloc(bitmap_len, 1);
-	if (read(dd->dfd, dd->bitmap, bitmap_len) < bitmap_len) {
-		error(INFO, "%s: cannot read memory bitmap\n",
-			DISKDUMP_VALID() ? "diskdump" : "compressed kdump");
-		goto err;
+
+	if (FLAT_FORMAT()) {
+		if (!read_flattened_format(dd->dfd, offset, dd->bitmap, bitmap_len)) {
+			error(INFO, "%s: cannot read memory bitmap\n",
+				DISKDUMP_VALID() ? "diskdump" : "compressed kdump");
+			goto err;
+		}
+	} else {
+		if (lseek(dd->dfd, offset, SEEK_SET) == failed) {
+			error(INFO, "%s: cannot lseek memory bitmap\n",
+				DISKDUMP_VALID() ? "diskdump" : "compressed kdump");
+			goto err;
+		}
+		if (read(dd->dfd, dd->bitmap, bitmap_len) < bitmap_len) {
+			error(INFO, "%s: cannot read memory bitmap\n",
+				DISKDUMP_VALID() ? "diskdump" : "compressed kdump");
+			goto err;
+		}
 	}
 
 	if (dump_is_partial(header))
@@ -348,6 +470,8 @@ restart:
 		dd->machine_type = EM_IA_64;
 	else if (machine_type("PPC64"))
 		dd->machine_type = EM_PPC64;
+	else if (machine_type("S390X"))
+		dd->machine_type = EM_S390;
 	else {
 		error(INFO, "%s: unsupported machine type: %s\n", 
 			DISKDUMP_VALID() ? "diskdump" : "compressed kdump",
@@ -355,6 +479,43 @@ restart:
 		goto err;
 	}
 
+	/* process elf notes data */
+	if (KDUMP_CMPRS_VALID() && !(dd->flags & NO_ELF_NOTES) &&
+		(dd->header->header_version >= 4) &&
+		(sub_header_kdump->offset_note) &&
+		(sub_header_kdump->size_note) && (machdep->process_elf_notes)) {
+		size = sub_header_kdump->size_note;
+		offset = sub_header_kdump->offset_note;
+
+		if ((dd->notes_buf = malloc(size)) == NULL)
+			error(FATAL, "compressed kdump: cannot malloc notes"
+				" buffer\n");
+
+		if ((dd->nt_prstatus_percpu = malloc(NR_CPUS * sizeof(void*))) == NULL)
+			error(FATAL, "compressed kdump: cannot malloc pointer"
+				" to NT_PRSTATUS notes\n");
+
+		if (FLAT_FORMAT()) {
+			if (!read_flattened_format(dd->dfd, offset, dd->notes_buf, size)) {
+				error(INFO, "compressed kdump: cannot read notes data"
+					"\n");
+				goto err;
+			}
+		} else {
+			if (lseek(dd->dfd, offset, SEEK_SET) == failed) {
+				error(INFO, "compressed kdump: cannot lseek notes data\n");
+				goto err;
+			}
+			if (read(dd->dfd, dd->notes_buf, size) < size) {
+				error(INFO, "compressed kdump: cannot read notes data"
+					"\n");
+				goto err;
+			}
+		}
+
+		machdep->process_elf_notes(dd->notes_buf, size);
+	}
+
 	/* For split dumpfile */
 	if (KDUMP_CMPRS_VALID()) {
 		is_split = ((dd->header->header_version >= 2) &&
@@ -375,6 +536,7 @@ restart:
 	if (!is_split) {
 		max_sect_len = divideup(header->max_mapnr, BITMAP_SECT_LEN);
 		pfn = 0;
+		dd->filename = file;
 	}
 	else {
 		ulong start = sub_header_kdump->start_pfn;
@@ -403,7 +565,13 @@ err:
 		free(dd->bitmap);
 	if (dd->dumpable_bitmap)
 		free(dd->dumpable_bitmap);
+	if (dd->notes_buf)
+		free(dd->notes_buf);
+	if (dd->nt_prstatus_percpu)
+		free(dd->nt_prstatus_percpu);
+
 	dd->flags &= ~(DISKDUMP_LOCAL|KDUMP_CMPRS_LOCAL);
+	pc->flags2 &= ~ELF_NOTES;
 	return FALSE;
 }
 
@@ -461,6 +629,9 @@ is_diskdump(char *file)
 	if (CRASHDEBUG(1))
 		__diskdump_memory_dump(fp);
 
+	if (pc->flags2 & GET_OSRELEASE) 
+		diskdump_get_osrelease();
+
 	return TRUE;
 }
 
@@ -588,22 +759,32 @@ cache_page(physaddr_t paddr)
 	desc_pos = pfn_to_pos(pfn);
 	seek_offset = dd->data_offset
 			+ (off_t)(desc_pos - 1)*sizeof(page_desc_t);
-	lseek(dd->dfd, seek_offset, SEEK_SET);
 
 	/* read page descriptor */
-	if (read(dd->dfd, &pd, sizeof(pd)) != sizeof(pd))
-		return READ_ERROR;
+	if (FLAT_FORMAT()) {
+		if (!read_flattened_format(dd->dfd, seek_offset, &pd, sizeof(pd)))
+			return READ_ERROR;
+	} else {
+		if (lseek(dd->dfd, seek_offset, SEEK_SET) == failed)
+			return SEEK_ERROR;
+		if (read(dd->dfd, &pd, sizeof(pd)) != sizeof(pd))
+			return READ_ERROR;
+	}
 
 	/* sanity check */
 	if (pd.size > block_size)
 		return READ_ERROR;
 
-	if (lseek(dd->dfd, pd.offset, SEEK_SET) == failed)
-		return SEEK_ERROR;
-
 	/* read page data */
-	if (read(dd->dfd, dd->compressed_page, pd.size) != pd.size)
-		return READ_ERROR;
+	if (FLAT_FORMAT()) {
+		if (!read_flattened_format(dd->dfd, pd.offset, dd->compressed_page, pd.size))
+			return READ_ERROR;
+	} else {
+		if (lseek(dd->dfd, pd.offset, SEEK_SET) == failed)
+			return SEEK_ERROR;
+		if (read(dd->dfd, dd->compressed_page, pd.size) != pd.size)
+			return READ_ERROR;
+	}
 
 	if (pd.flags & DUMP_DH_COMPRESSED) {
 		retlen = block_size;
@@ -699,7 +880,7 @@ get_diskdump_panic_task(void)
 	return (ulong)dd->header->tasks[dd->header->current_cpu];
 }
 
-extern  void get_netdump_regs_x86(struct bt_info *, ulong *, ulong *);
+extern void get_netdump_regs_x86(struct bt_info *, ulong *, ulong *);
 extern void get_netdump_regs_x86_64(struct bt_info *, ulong *, ulong *);
 
 static void
@@ -750,6 +931,9 @@ get_diskdump_regs(struct bt_info *bt, ulong *eip, ulong *esp)
 	case EM_X86_64:
 		return get_netdump_regs_x86_64(bt, eip, esp);
 		break;
+	case EM_S390:
+		return machdep->get_stack_frame(bt, eip, esp);
+		break;
 
 	default:
 		error(FATAL, "%s: unsupported machine type: %s\n",
@@ -781,11 +965,101 @@ diskdump_free_memory(void)
         return 0;
 }
 
-int diskdump_memory_used(void)
+int 
+diskdump_memory_used(void)
 {
         return 0;
 }
 
+static void 
+dump_vmcoreinfo(FILE *fp)
+{
+	char *buf = NULL;
+	unsigned long i = 0;
+	unsigned long size_vmcoreinfo = dd->sub_header_kdump->size_vmcoreinfo;
+	off_t offset = dd->sub_header_kdump->offset_vmcoreinfo;
+	const off_t failed = (off_t)-1;
+
+	if ((buf = malloc(size_vmcoreinfo)) == NULL) {
+		error(FATAL, "compressed kdump: cannot malloc vmcoreinfo"
+				" buffer\n");
+	}
+
+	if (FLAT_FORMAT()) {
+		if (!read_flattened_format(dd->dfd, offset, buf, size_vmcoreinfo)) {
+			error(INFO, "compressed kdump: cannot read vmcoreinfo data\n");
+			goto err;
+		}
+	} else {
+		if (lseek(dd->dfd, offset, SEEK_SET) == failed) {
+			error(INFO, "compressed kdump: cannot lseek dump vmcoreinfo\n");
+			goto err;
+		}
+		if (read(dd->dfd, buf, size_vmcoreinfo) < size_vmcoreinfo) {
+			error(INFO, "compressed kdump: cannot read vmcoreinfo data\n");
+			goto err;
+		}
+	}
+
+	fprintf(fp, "                      ");
+	for (i = 0; i < size_vmcoreinfo; i++) {
+		fprintf(fp, "%c", buf[i]);
+		if (buf[i] == '\n')
+			fprintf(fp, "                      ");
+	}
+	if (buf[i - 1] != '\n')
+		fprintf(fp, "\n");
+err:
+	if (buf)
+		free(buf);
+	return;
+}
+
+static void 
+dump_nt_prstatus_offset(FILE *fp)
+{
+	struct kdump_sub_header *sub_header_kdump = dd->sub_header_kdump;
+	size_t size;
+	off_t offset;
+	Elf32_Nhdr *note32 = NULL;
+	Elf64_Nhdr *note64 = NULL;
+	size_t tot, len = 0;
+
+	if (KDUMP_CMPRS_VALID() && !(dd->flags & NO_ELF_NOTES) &&
+	    (dd->header->header_version >= 4) &&
+	    (sub_header_kdump->offset_note) &&
+	    (sub_header_kdump->size_note) && (machdep->process_elf_notes)) {
+		size = sub_header_kdump->size_note;
+		offset = sub_header_kdump->offset_note;
+
+		fprintf(fp, "  NT_PRSTATUS_offset: ");
+		for (tot = 0; tot < size; tot += len) {
+			if (machine_type("X86_64") || machine_type("S390X")) {
+				note64 = (void *)dd->notes_buf + tot;
+				len = sizeof(Elf64_Nhdr);
+				len = roundup(len + note64->n_namesz, 4);
+				len = roundup(len + note64->n_descsz, 4);
+
+				if (note64->n_type == NT_PRSTATUS)
+					fprintf(fp, "%s%lx\n",
+						(tot == 0) ? "" : "                      ",
+						(ulong)(offset + tot));
+
+			} else if (machine_type("X86")) {
+				note32 = (void *)dd->notes_buf + tot;
+				len = sizeof(Elf32_Nhdr);
+				len = roundup(len + note32->n_namesz, 4);
+				len = roundup(len + note32->n_descsz, 4);
+
+				if (note32->n_type == NT_PRSTATUS)
+					fprintf(fp, "%s%lx\n",
+						(tot == 0) ? "" : "                      ",
+						(ulong)(offset + tot));
+			}
+		}
+	}
+}
+
 /*
  *  This function is dump-type independent, and could be used
  *  to dump the diskdump_data structure contents and perhaps
@@ -800,6 +1074,9 @@ __diskdump_memory_dump(FILE *fp)
 	struct kdump_sub_header *kdsh;
 	ulong *tasks;
 
+	if (FLAT_FORMAT())
+		dump_flat_header(fp);
+
         fprintf(fp, "diskdump_data: \n");
 	fprintf(fp, "          filename: %s\n", dd->filename);
         fprintf(fp, "             flags: %lx (", dd->flags);
@@ -812,7 +1089,9 @@ __diskdump_memory_dump(FILE *fp)
                 fprintf(fp, "%sERROR_EXCLUDED", others++ ? "|" : "");
         if (dd->flags & ZERO_EXCLUDED)
                 fprintf(fp, "%sZERO_EXCLUDED", others++ ? "|" : "");
-        fprintf(fp, ")\n");
+	if (dd->flags & NO_ELF_NOTES)
+		fprintf(fp, "%sNO_ELF_NOTES", others++ ? "|" : "");
+        fprintf(fp, ") %s\n", FLAT_FORMAT() ? "[FLAT]" : "");
         fprintf(fp, "               dfd: %d\n", dd->dfd);
         fprintf(fp, "               ofp: %lx\n", (ulong)dd->ofp);
         fprintf(fp, "      machine_type: %d ", dd->machine_type);
@@ -828,6 +1107,8 @@ __diskdump_memory_dump(FILE *fp)
 		fprintf(fp, "(EM_IA_64)\n"); break;
 	case EM_PPC64:
 		fprintf(fp, "(EM_PPC64)\n"); break;
+	case EM_S390:
+		fprintf(fp, "(EM_S390)\n"); break;
 	default:
 		fprintf(fp, "(unknown)\n"); break;
 	}
@@ -954,6 +1235,31 @@ __diskdump_memory_dump(FILE *fp)
 			fprintf(fp, "           start_pfn: %lu\n", dd->sub_header_kdump->start_pfn);
 			fprintf(fp, "             end_pfn: %lu\n", dd->sub_header_kdump->end_pfn);
 		}
+		if (dh->header_version >= 3) {
+			fprintf(fp, "   offset_vmcoreinfo: %lx\n",
+				(ulong)dd->sub_header_kdump->offset_vmcoreinfo);
+			fprintf(fp, "     size_vmcoreinfo: %lu\n",
+				dd->sub_header_kdump->size_vmcoreinfo);
+			if (dd->sub_header_kdump->offset_vmcoreinfo &&
+				dd->sub_header_kdump->size_vmcoreinfo) {
+				dump_vmcoreinfo(fp);
+			}
+		}
+		if (dh->header_version >= 4) {
+			fprintf(fp, "         offset_note: %lx\n",
+				(ulong)dd->sub_header_kdump->offset_note);
+			fprintf(fp, "           size_note: %lu\n",
+				dd->sub_header_kdump->size_note);
+			fprintf(fp, "  num_prstatus_notes: %d\n",
+				dd->num_prstatus_notes);
+			fprintf(fp, "           notes_buf: %lx\n",
+				(ulong)dd->notes_buf);
+			for (i = 0; i < dd->num_prstatus_notes; i++) {
+				fprintf(fp, "            notes[%d]: %lx\n",
+					i, (ulong)dd->nt_prstatus_percpu[i]);
+			}
+			dump_nt_prstatus_offset(fp);
+		}
 		fprintf(fp, "\n");
 	} else
         	fprintf(fp, "(n/a)\n\n");
@@ -1078,3 +1384,82 @@ show_split_dumpfiles(void)
 			fprintf(fp, "\n");
 	}
 }
+
+void *
+diskdump_get_prstatus_percpu(int cpu)
+{
+	return dd->nt_prstatus_percpu[cpu];
+}
+
+/*
+ * Reads a string value from VMCOREINFO.
+ *
+ * Returns a string (that has to be freed by the caller) that contains the
+ * value for key or NULL if the key has not been found.
+ */
+static char *
+vmcoreinfo_read_string(const char *key)
+{
+	char *buf, *value_string, *p1, *p2;
+	size_t value_length;
+	ulong size_vmcoreinfo;
+	off_t offset;
+	char keybuf[BUFSIZE];
+	const off_t failed = (off_t)-1;
+
+	buf = value_string = NULL;
+	size_vmcoreinfo = dd->sub_header_kdump->size_vmcoreinfo;
+	offset = dd->sub_header_kdump->offset_vmcoreinfo;
+	sprintf(keybuf, "%s=", key);
+
+	if ((buf = malloc(size_vmcoreinfo+1)) == NULL) {
+		error(INFO, "compressed kdump: cannot malloc vmcoreinfo"
+			    " buffer\n");
+		goto err;
+	}
+
+	if (FLAT_FORMAT()) {
+		if (!read_flattened_format(dd->dfd, offset, buf, size_vmcoreinfo)) {
+			error(INFO, "compressed kdump: cannot read vmcoreinfo data\n");
+			goto err;
+		}
+	} else {
+		if (lseek(dd->dfd, offset, SEEK_SET) == failed) {
+			error(INFO, "compressed kdump: cannot lseek dump vmcoreinfo\n");
+			goto err;
+		}
+		if (read(dd->dfd, buf, size_vmcoreinfo) < size_vmcoreinfo) {
+			error(INFO, "compressed kdump: cannot read vmcoreinfo data\n");
+			goto err;
+		}
+	}
+
+	buf[size_vmcoreinfo] = '\n';
+
+	if ((p1 = strstr(buf, keybuf))) {
+		p2 = p1 + strlen(keybuf);
+		p1 = strstr(p2, "\n");
+		value_length = p1-p2;
+		value_string = calloc(value_length+1, sizeof(char));
+		strncpy(value_string, p2, value_length);
+		value_string[value_length] = NULLCHAR;
+	}
+err:
+	if (buf)
+		free(buf);
+
+	return value_string;
+}
+
+static void
+diskdump_get_osrelease(void)
+{
+	char *string;
+
+	if ((string = vmcoreinfo_read_string("OSRELEASE"))) {
+		fprintf(fp, "%s\n", string);
+		free(string);
+	}
+	else
+		pc->flags2 &= ~GET_OSRELEASE;
+}
