@@ -1,8 +1,8 @@
 /* main.c - core analysis suite
  *
  * Copyright (C) 1999, 2000, 2001, 2002 Mission Critical Linux, Inc.
- * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010 David Anderson
- * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010 Red Hat, Inc. All rights reserved.
+ * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 David Anderson
+ * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Red Hat, Inc. All rights reserved.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -26,6 +26,7 @@ static int is_builtin_command(void);
 static int is_input_file(void);
 static void check_xen_hyper(void);
 static void show_untrusted_files(void);
+static void get_osrelease(char *);
 
 static struct option long_options[] = {
         {"memory_module", required_argument, 0, 0},
@@ -60,6 +61,10 @@ static struct option long_options[] = {
 	{"active", 0, 0, 0},
 	{"minimal", 0, 0, 0},
 	{"mod", required_argument, 0, 0},
+	{"kvmhost", required_argument, 0, 0},
+	{"kvmio", required_argument, 0, 0},
+	{"no_elf_notes", 0, 0, 0},
+	{"osrelease", required_argument, 0, 0},
         {0, 0, 0, 0}
 };
 
@@ -67,6 +72,7 @@ int
 main(int argc, char **argv)
 {
 	int i, c, option_index;
+	char *tmpname;
 
 	setup_environment(argc, argv);
 
@@ -174,6 +180,15 @@ main(int argc, char **argv)
 		        else if (STREQ(long_options[option_index].name, "zero_excluded")) 
 				*diskdump_flags |= ZERO_EXCLUDED;
 
+			else if (STREQ(long_options[option_index].name, "no_elf_notes")) {
+				if (machine_type("X86") || machine_type("X86_64"))
+					*diskdump_flags |= NO_ELF_NOTES;
+				else
+					error(INFO,
+					      "--no_elf_notes is only applicable to "
+					      "the X86 and X86_64 architectures.\n");
+			}
+
 		        else if (STREQ(long_options[option_index].name, "no_panic")) 
 				tt->flags |= PANIC_TASK_NOT_FOUND;
 
@@ -219,6 +234,17 @@ main(int argc, char **argv)
 			else if (STREQ(long_options[option_index].name, "minimal")) 
 				pc->flags |= MINIMAL_MODE;
 
+		        else if (STREQ(long_options[option_index].name, "kvmhost"))
+				set_kvmhost_type(optarg);
+
+		        else if (STREQ(long_options[option_index].name, "kvmio"))
+				set_kvm_iohole(optarg);
+
+		        else if (STREQ(long_options[option_index].name, "osrelease")) {
+				pc->flags2 |= GET_OSRELEASE;
+				get_osrelease(optarg);
+			}
+
 			else {
 				error(INFO, "internal error: option %s unhandled\n",
 					long_options[option_index].name);
@@ -351,6 +377,12 @@ main(int argc, char **argv)
 			continue;
 		}
 
+		if (STREQ(argv[optind], "/dev/crash")) {
+			pc->memory_device = argv[optind];
+			optind++;
+			continue;
+		}
+
        		if (!file_exists(argv[optind], NULL)) {
                 	error(INFO, "%s: %s\n", argv[optind], strerror(ENOENT));
                 	program_usage(SHORT_FORM);
@@ -367,8 +399,30 @@ main(int argc, char **argv)
 			} else
 				pc->namelist = argv[optind];
 
+		} else if (is_compressed_kernel(argv[optind], &tmpname)) {
+			if (pc->namelist) {
+				if (!select_namelist(tmpname)) {
+					error(INFO, 
+					    "too many namelist arguments\n");
+					program_usage(SHORT_FORM);
+				}
+				if (pc->namelist_debug == tmpname) {
+					pc->namelist_debug_orig = argv[optind];
+				} else {
+					pc->namelist_debug_orig = pc->namelist_orig;
+					pc->namelist_orig = argv[optind];
+				}
+			} else {
+				pc->namelist = tmpname;
+				pc->namelist_orig = argv[optind];
+			}
+			pc->cleanup = NULL;
+
 		} else if (!(pc->flags & KERNEL_DEBUG_QUERY)) {
 
+			if (is_flattened_format(argv[optind]))
+				pc->flags2 |= FLAT;
+
 			if (STREQ(argv[optind], "/dev/mem")) {
                         	if (pc->flags & MEMORY_SOURCES) {
                                 	error(INFO, 
@@ -381,6 +435,18 @@ main(int argc, char **argv)
 				pc->writemem = write_dev_mem;
 				pc->live_memsrc = argv[optind];
 
+			} else if (is_proc_kcore(argv[optind], KCORE_LOCAL)) {
+				if (pc->flags & MEMORY_SOURCES) {
+					error(INFO, 
+					    "too many dumpfile arguments\n");
+					program_usage(SHORT_FORM);
+				}
+				pc->flags |= PROC_KCORE;
+				pc->dumpfile = NULL;
+				pc->readmem = read_proc_kcore;
+				pc->writemem = write_proc_kcore;
+				pc->live_memsrc = argv[optind];
+
 			} else if (is_netdump(argv[optind], NETDUMP_LOCAL)) {
                                 if (pc->flags & MEMORY_SOURCES) {
                                         error(INFO,
@@ -856,10 +922,12 @@ setup_environment(int argc, char **argv)
 	pc->memory_device = MEMORY_DRIVER_DEVICE;
 	machdep->bits = sizeof(long) * 8;
 	machdep->verify_paddr = generic_verify_paddr;
+	machdep->get_kvaddr_ranges = generic_get_kvaddr_ranges;
 	pc->redhat_debug_loc = DEFAULT_REDHAT_DEBUG_LOCATION;
 	pc->cmdgencur = 0;
 	pc->cmd_table = linux_command_table;
 	kt->BUG_bytes = -1;
+	kt->flags |= PRE_KERNEL_INIT;
 
 	/*
 	 *  Get gdb version before initializing it since this might be one 
@@ -1147,6 +1215,9 @@ dump_program_context(void)
         if (pc->flags & PRELOAD_EXTENSIONS)
                 sprintf(&buf[strlen(buf)], 
 			"%sPRELOAD_EXTENSIONS", others++ ? "|" : "");
+        if (pc->flags & PROC_KCORE)
+                sprintf(&buf[strlen(buf)], 
+			"%sPROC_KCORE", others++ ? "|" : "");
 
 	if (pc->flags)
 		strcat(buf, ")");
@@ -1172,12 +1243,24 @@ dump_program_context(void)
 			fprintf(fp, "%s\n", buf);
 	}
 
+	others = 0;
+	fprintf(fp, "           flags2: %llx (", pc->flags2);
+	if (pc->flags2 & FLAT)
+		fprintf(fp, "%sFLAT", others++ ? "|" : "");
+	if (pc->flags2 & ELF_NOTES)
+		fprintf(fp, "%sELF_NOTES", others++ ? "|" : "");
+	if (pc->flags2 & GET_OSRELEASE)
+		fprintf(fp, "%sGET_OSRELEASE", others++ ? "|" : "");
+	fprintf(fp, ")\n");
+
 	fprintf(fp, "         namelist: %s\n", pc->namelist);
 	fprintf(fp, "         dumpfile: %s\n", pc->dumpfile);
 	fprintf(fp, "      live_memsrc: %s\n", pc->live_memsrc);
 	fprintf(fp, "       system_map: %s\n", pc->system_map);
 	fprintf(fp, "   namelist_debug: %s\n", pc->namelist_debug);
 	fprintf(fp, "   debuginfo_file: %s\n", pc->debuginfo_file);
+	fprintf(fp, "    namelist_orig: %s\n", pc->namelist_orig);
+	fprintf(fp, "namelist_dbg_orig: %s\n", pc->namelist_debug_orig);
 	fprintf(fp, "  kvmdump_mapfile: %s\n", pc->kvmdump_mapfile);
 	fprintf(fp, "    memory_module: %s\n", pc->memory_module);
 	fprintf(fp, "    memory_device: %s\n", pc->memory_device);
@@ -1348,6 +1431,10 @@ dump_program_context(void)
 		fprintf(fp, "%sMODULE_TREE", others ? "|" : "");
         if (pc->curcmd_flags & IGNORE_ERRORS)
 		fprintf(fp, "%sIGNORE_ERRORS", others ? "|" : "");
+        if (pc->curcmd_flags & FROM_RCFILE)
+		fprintf(fp, "%sFROM_RCFILE", others ? "|" : "");
+        if (pc->curcmd_flags & MEMTYPE_KVADDR)
+		fprintf(fp, "%sMEMTYPE_KVADDR", others ? "|" : "");
 	fprintf(fp, ")\n");
 	fprintf(fp, "   curcmd_private: %llx\n", pc->curcmd_private); 
 	fprintf(fp, "       sigint_cnt: %d\n", pc->sigint_cnt);
@@ -1421,6 +1508,7 @@ dump_program_context(void)
 		dumpfile_memory(DUMPFILE_MEM_USED)); 
 	fprintf(fp, "           curext: %lx\n", (ulong)pc->curext); 
 	fprintf(fp, "             sbrk: %lx\n", (ulong)pc->sbrk); 
+	fprintf(fp, "          cleanup: %s\n", pc->cleanup);
 }
 
 /*
@@ -1445,6 +1533,13 @@ clean_exit(int status)
 	if (pc->flags & MEMMOD)
 		cleanup_memory_driver();
 
+	if ((pc->namelist_orig) && file_exists(pc->namelist, NULL))
+		unlink(pc->namelist);
+	if ((pc->namelist_debug_orig) && file_exists(pc->namelist_debug, NULL))
+		unlink(pc->namelist_debug);
+	if (pc->cleanup && file_exists(pc->cleanup, NULL))
+		unlink(pc->cleanup);
+
 	exit(status);
 }
 
@@ -1515,3 +1610,29 @@ show_untrusted_files(void)
 	if (cnt)
 		fprintf(fp, "\n");
 }
+
+/*
+ *  If GET_OSRELEASE is still set, the OS release has been
+ *  found and displayed.
+ */
+static void
+get_osrelease(char *dumpfile)
+{
+	int retval = 1;
+
+	if (is_flattened_format(dumpfile))
+		pc->flags2 |= FLAT;
+	
+	if (is_diskdump(dumpfile)) {
+		if (pc->flags2 & GET_OSRELEASE)
+			retval = 0;
+	} else if (is_kdump(dumpfile, KDUMP_LOCAL)) {
+		if (pc->flags2 & GET_OSRELEASE)
+			retval = 0;
+	}
+	
+	if (retval)
+		fprintf(fp, "unknown\n");
+
+	clean_exit(retval);
+}
