@@ -1,8 +1,8 @@
 /* net.c - core analysis suite
  *
  * Copyright (C) 1999, 2000, 2001, 2002 Mission Critical Linux, Inc.
- * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010 David Anderson
- * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010 Red Hat, Inc. All rights reserved.
+ * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 David Anderson
+ * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Red Hat, Inc. All rights reserved.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -204,11 +204,19 @@ net_init(void)
 				MEMBER_OFFSET_INIT(inet_opt_num, "inet_opt", "num");
 			} else {	/* inet_opt moved to inet_sock */
 				ASSIGN_OFFSET(inet_sock_inet) = 0;
-				MEMBER_OFFSET_INIT(inet_opt_daddr, "inet_sock", "daddr");
-				MEMBER_OFFSET_INIT(inet_opt_rcv_saddr, "inet_sock", "rcv_saddr");
-				MEMBER_OFFSET_INIT(inet_opt_dport, "inet_sock", "dport");
-				MEMBER_OFFSET_INIT(inet_opt_sport, "inet_sock", "sport");
-				MEMBER_OFFSET_INIT(inet_opt_num, "inet_sock", "num");
+				if (MEMBER_EXISTS("inet_sock", "daddr")) {
+					MEMBER_OFFSET_INIT(inet_opt_daddr, "inet_sock", "daddr");
+					MEMBER_OFFSET_INIT(inet_opt_rcv_saddr, "inet_sock", "rcv_saddr");
+					MEMBER_OFFSET_INIT(inet_opt_dport, "inet_sock", "dport");
+					MEMBER_OFFSET_INIT(inet_opt_sport, "inet_sock", "sport");
+					MEMBER_OFFSET_INIT(inet_opt_num, "inet_sock", "num");
+				} else {
+					MEMBER_OFFSET_INIT(inet_opt_daddr, "inet_sock", "inet_daddr");
+					MEMBER_OFFSET_INIT(inet_opt_rcv_saddr, "inet_sock", "inet_rcv_saddr");
+					MEMBER_OFFSET_INIT(inet_opt_dport, "inet_sock", "inet_dport");
+					MEMBER_OFFSET_INIT(inet_opt_sport, "inet_sock", "inet_sport");
+					MEMBER_OFFSET_INIT(inet_opt_num, "inet_sock", "inet_num");
+				}
 			}	
 
 			if (VALID_STRUCT(inet_sock) && 
@@ -778,7 +786,8 @@ get_sock_info(ulong sock, char *buf)
 {
 	uint32_t daddr, rcv_saddr;
 	uint16_t dport, sport;
-	ushort num, family, type;
+	ushort family, type;
+	ushort num ATTRIBUTE_UNUSED;
 	char *sockbuf, *inet_sockbuf;
 	ulong ipv6_pinfo, ipv6_rcv_saddr, ipv6_daddr;
 	uint16_t u6_addr16_src[8];
