@@ -1,7 +1,7 @@
 /* netdump.c 
  *
- * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010 David Anderson
- * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010 Red Hat, Inc. All rights reserved.
+ * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 David Anderson
+ * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Red Hat, Inc. All rights reserved.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,6 +24,8 @@
 static struct vmcore_data vmcore_data = { 0 };
 static struct vmcore_data *nd = &vmcore_data;
 static struct xen_kdump_data xen_kdump_data = { 0 };
+static struct proc_kcore_data proc_kcore_data = { 0 };
+static struct proc_kcore_data *pkd = &proc_kcore_data;
 static void netdump_print(char *, ...);
 static void dump_Elf32_Ehdr(Elf32_Ehdr *);
 static void dump_Elf32_Phdr(Elf32_Phdr *, int);
@@ -35,6 +37,10 @@ static void get_netdump_regs_ppc64(struct bt_info *, ulong *, ulong *);
 static void get_netdump_regs_arm(struct bt_info *, ulong *, ulong *);
 static physaddr_t xen_kdump_p2m(physaddr_t);
 static void check_dumpfile_size(char *);
+static int proc_kcore_init_32(FILE *fp);
+static int proc_kcore_init_64(FILE *fp);
+static char *get_regs_from_note(char *, ulong *, ulong *);
+static void kdump_get_osrelease(void);
 
 #define ELFSTORE 1
 #define ELFREAD  0
@@ -117,16 +123,16 @@ is_netdump(char *file, ulong source_query)
 	}
 
 	size = MIN_NETDUMP_ELF_HEADER_SIZE;
-        if (read(fd, eheader, size) != size) {
-                sprintf(buf, "%s: read", file);
-                perror(buf);
-		goto bailout;
-	}
 
-        if (lseek(fd, 0, SEEK_SET) != 0) {
-                sprintf(buf, "%s: lseek", file);
-                perror(buf);
-                goto bailout;
+	if (FLAT_FORMAT()) {
+		if (!read_flattened_format(fd, 0, eheader, size))
+			goto bailout;
+	} else {
+		if (read(fd, eheader, size) != size) {
+			sprintf(buf, "%s: read", file);
+			perror(buf);
+			goto bailout;
+		}
 	}
 
 	load32 = NULL;
@@ -259,6 +265,11 @@ is_netdump(char *file, ulong source_query)
 		goto bailout;
 	}
 
+	if (source_query == KCORE_LOCAL) {
+		close(fd);
+		return TRUE;
+	}
+
 	switch (DUMPFILE_FORMAT(tmp_flags))
 	{
 	case NETDUMP_ELF32:
@@ -281,12 +292,24 @@ is_netdump(char *file, ulong source_query)
 		clean_exit(1);
 	}
 
-        if (read(fd, tmp_elf_header, size) != size) {
-                sprintf(buf, "%s: read", file);
-                perror(buf);
-		free(tmp_elf_header);
-                goto bailout;
-        }
+	if (FLAT_FORMAT()) {
+		if (!read_flattened_format(fd, 0, tmp_elf_header, size)) {
+			free(tmp_elf_header);
+			goto bailout;
+		}
+	} else {
+		if (lseek(fd, 0, SEEK_SET) != 0) {
+			sprintf(buf, "%s: lseek", file);
+			perror(buf);
+			goto bailout;
+		}
+		if (read(fd, tmp_elf_header, size) != size) {
+			sprintf(buf, "%s: read", file);
+			perror(buf);
+			free(tmp_elf_header);
+			goto bailout;
+		}
+	}
 
 	nd->ndfd = fd;
 	nd->elf_header = tmp_elf_header;
@@ -357,6 +380,10 @@ is_netdump(char *file, ulong source_query)
 	if (CRASHDEBUG(1))
 		netdump_memory_dump(fp);
 
+	if ((source_query == KDUMP_LOCAL) && 
+	    (pc->flags2 & GET_OSRELEASE))
+		kdump_get_osrelease();
+
 	return nd->header_size;
 
 bailout:
@@ -514,11 +541,15 @@ read_netdump(int fd, void *bufptr, int cnt, ulong addr, physaddr_t paddr)
 		break;
 	}	
 
-        if (lseek(nd->ndfd, offset, SEEK_SET) == -1)
-                return SEEK_ERROR;
-
-        if (read(nd->ndfd, bufptr, cnt) != cnt)
-                return READ_ERROR;
+	if (FLAT_FORMAT()) {
+		if (!read_flattened_format(nd->ndfd, offset, bufptr, cnt))
+			return READ_ERROR;
+	} else {
+		if (lseek(nd->ndfd, offset, SEEK_SET) == -1)
+			return SEEK_ERROR;
+		if (read(nd->ndfd, bufptr, cnt) != cnt)
+			return READ_ERROR;
+	}
 
         return cnt;
 }
@@ -845,6 +876,9 @@ netdump_memory_dump(FILE *fp)
 	fpsave = nd->ofp;
 	nd->ofp = fp;
 
+	if (FLAT_FORMAT())
+		dump_flat_header(nd->ofp);
+
 	netdump_print("vmcore_data: \n");
 	netdump_print("                  flags: %lx (", nd->flags);
 	others = 0;
@@ -864,7 +898,7 @@ netdump_memory_dump(FILE *fp)
 		netdump_print("%sKDUMP_ELF64", others++ ? "|" : "");
 	if (nd->flags & PARTIAL_DUMP)
 		netdump_print("%sPARTIAL_DUMP", others++ ? "|" : "");
-	netdump_print(")\n");
+	netdump_print(") %s\n", FLAT_FORMAT() ? "[FLAT]" : "");
 	if ((pc->flags & RUNTIME) && symbol_exists("dump_level")) {
 		int dump_level;
                 if (readmem(symbol_value("dump_level"), KVADDR, &dump_level,
@@ -1294,6 +1328,9 @@ dump_Elf64_Ehdr(Elf64_Ehdr *elf)
         case EM_X86_64:
                 netdump_print("(EM_X86_64)\n");
                 break;
+	case EM_S390:
+                netdump_print("(EM_S390)\n");
+                break;
         default:
                 netdump_print("(unsupported)\n");
                 break;
@@ -1557,7 +1594,7 @@ vmcoreinfo_read_string(const char *key)
 			}
 
 			value_length = end - (1+ i + key_length);
-			value = malloc(value_length);
+			value = calloc(value_length+1, sizeof(char));
 			if (value)
 				strncpy(value, vmcoreinfo + i + key_length + 1, 
 					value_length);
@@ -1698,9 +1735,11 @@ dump_Elf32_Nhdr(Elf32_Off offset, int store)
 			if (store)
 				error(WARNING, "unknown Xen n_type: %lx\n\n", 
 					note->n_type);
-		} else if (vmcoreinfo)
+		} else if (vmcoreinfo) {
 			netdump_print("(unused)\n");
-		else
+			nd->vmcoreinfo = (char *)(ptr + note->n_namesz + 1);
+			nd->size_vmcoreinfo = note->n_descsz;
+		} else
 			netdump_print("(?)\n");
 		break;
 
@@ -1876,6 +1915,24 @@ dump_Elf64_Nhdr(Elf64_Off offset, int store)
 		if (store)
 			nd->nt_prpsinfo = (void *)note;
 		break;
+	case NT_FPREGSET:
+		netdump_print("(NT_FPREGSET)\n");
+		break;
+	case NT_S390_TIMER:
+		netdump_print("(NT_S390_TIMER)\n");
+		break;
+	case NT_S390_TODCMP:
+		netdump_print("(NT_S390_TODCMP)\n");
+		break;
+	case NT_S390_TODPREG:
+		netdump_print("(NT_S390_TODPREG)\n");
+		break;
+	case NT_S390_CTRS:
+		netdump_print("(NT_S390_CTRS)\n");
+		break;
+	case NT_S390_PREFIX:
+		netdump_print("(NT_S390_PREFIX)\n");
+		break;
 	case NT_TASKSTRUCT:
 		netdump_print("(NT_TASKSTRUCT)\n");
 		if (store) {
@@ -1932,14 +1989,14 @@ dump_Elf64_Nhdr(Elf64_Off offset, int store)
 		} else if (vmcoreinfo) {
                         netdump_print("(unused)\n");
 
-			if (READ_PAGESIZE_FROM_VMCOREINFO() && store) {
-				nd->vmcoreinfo = (char *)nd->elf64 + offset +
-					(sizeof(Elf64_Nhdr) +
-					((note->n_namesz + 3) & ~3));
-				nd->size_vmcoreinfo = note->n_descsz;
+			nd->vmcoreinfo = (char *)nd->elf64 + offset +
+				(sizeof(Elf64_Nhdr) +
+				((note->n_namesz + 3) & ~3));
+			nd->size_vmcoreinfo = note->n_descsz;
+
+			if (READ_PAGESIZE_FROM_VMCOREINFO() && store)
 				nd->page_size = (uint)
 					vmcoreinfo_read_integer("PAGESIZE", 0);
-			}
                 } else
                         netdump_print("(?)\n");
                 break;
@@ -2003,19 +2060,24 @@ dump_Elf64_Nhdr(Elf64_Off offset, int store)
                 break;
 	}
 
-	if (store && machine_type("S390X"))
-		machdep->dumpfile_init(nd->num_prstatus_notes, note);
+	if (machine_type("S390X")) {
+		if (store)
+			machdep->dumpfile_init(nd->num_prstatus_notes, note);
 
-	uptr = (ulonglong *)(ptr + note->n_namesz);
-
-        /*
-         * kdumps are off-by-1, because their n_namesz is 5 for "CORE".
-         */
-        if ((nd->flags & KDUMP_ELF64) && (note->n_namesz == 5))
-                uptr = (ulonglong *)(ptr + ((note->n_namesz + 3) & ~3));
-
-        if (xen_core)
-                uptr = (ulonglong *)roundup((ulong)uptr, 4);
+		uptr = (ulonglong *)
+	    	    ((void *)note + roundup(sizeof(*note) + note->n_namesz, 4));
+	} else {
+		uptr = (ulonglong *)(ptr + note->n_namesz);
+	
+		/*
+		 * kdumps are off-by-1, because their n_namesz is 5 for "CORE".
+		 */
+		if ((nd->flags & KDUMP_ELF64) && (note->n_namesz == 5))
+			uptr = (ulonglong *)(ptr + ((note->n_namesz + 3) & ~3));
+	
+		if (xen_core)
+			uptr = (ulonglong *)roundup((ulong)uptr, 4);
+	}
 
 	if (BITS32() && (xen_core || (note->n_type == NT_PRSTATUS))) {
 		iptr = (int *)uptr;
@@ -2037,6 +2099,10 @@ dump_Elf64_Nhdr(Elf64_Off offset, int store)
 				netdump_print("                         ");
 		}
 		lf = 0;
+	} else if (note->n_descsz == 4) {
+		i = 0; lf = 1;
+		iptr = (int *)uptr;
+		netdump_print("                         %08lx\n", *iptr); 
 	} else {
 		for (i = lf = 0; i < note->n_descsz/sizeof(ulonglong); i++) {
 			if (((i%2)==0)) {
@@ -2112,6 +2178,42 @@ get_netdump_regs(struct bt_info *bt, ulong *eip, ulong *esp)
 	}
 }
 
+/* 
+ * get regs from elf note, and return the address of user_regs. 
+ */
+static char * 
+get_regs_from_note(char *note, ulong *ip, ulong *sp)
+{
+	Elf32_Nhdr *note32;
+	Elf64_Nhdr *note64;
+	size_t len;
+	char *user_regs;
+	long offset_sp, offset_ip;
+
+	if (machine_type("X86_64")) {
+		note64 = (Elf64_Nhdr *)note;
+		len = sizeof(Elf64_Nhdr);
+		len = roundup(len + note64->n_namesz, 4);
+		len = roundup(len + note64->n_descsz, 4);
+		offset_sp = OFFSET(user_regs_struct_rsp);
+		offset_ip = OFFSET(user_regs_struct_rip);
+	} else if (machine_type("X86")) {
+		note32 = (Elf32_Nhdr *)note;
+		len = sizeof(Elf32_Nhdr);
+		len = roundup(len + note32->n_namesz, 4);
+		len = roundup(len + note32->n_descsz, 4);
+		offset_sp = OFFSET(user_regs_struct_esp);
+		offset_ip = OFFSET(user_regs_struct_eip);
+	} else
+		return NULL;
+
+	user_regs = note + len - SIZE(user_regs_struct) - sizeof(long);
+	*sp = ULONG(user_regs + offset_sp);
+	*ip = ULONG(user_regs + offset_ip);
+
+	return user_regs;
+}
+
 struct x86_64_user_regs_struct {
         unsigned long r15,r14,r13,r12,rbp,rbx,r11,r10;
         unsigned long r9,r8,rax,rcx,rdx,rsi,rdi,orig_rax;
@@ -2128,6 +2230,7 @@ get_netdump_regs_x86_64(struct bt_info *bt, ulong *ripp, ulong *rspp)
         size_t len;
         char *user_regs;
 	ulong regs_size, rsp_offset, rip_offset;
+	ulong rip, rsp;
 
         if (is_task_active(bt->task)) 
                 bt->flags |= BT_DUMPFILE_SEARCH;
@@ -2168,12 +2271,31 @@ get_netdump_regs_x86_64(struct bt_info *bt, ulong *ripp, ulong *rspp)
 			*ripp = ULONG(user_regs + rip_offset);
 
 			if (*ripp && *rspp)
-				return;
+				bt->flags |= BT_KDUMP_ELF_REGS;
 		}
 			
 		bt->machdep = (void *)user_regs;
 	}
 
+	if (ELF_NOTES_VALID() && 
+	    (bt->flags & BT_DUMPFILE_SEARCH) && DISKDUMP_DUMPFILE() && 
+	    (note = (Elf64_Nhdr *)
+	     diskdump_get_prstatus_percpu(bt->tc->processor))) {
+		user_regs = get_regs_from_note((char *)note, &rip, &rsp);
+
+		if (CRASHDEBUG(1))
+			netdump_print("ELF prstatus rsp: %lx rip: %lx\n",
+				rsp, rip);
+
+		*rspp = rsp;
+		*ripp = rip;
+
+		if (*ripp && *rspp)
+			bt->flags |= BT_KDUMP_ELF_REGS;
+
+		bt->machdep = (void *)user_regs;
+	}
+
         machdep->get_stack_frame(bt, ripp, rspp);
 }
 
@@ -2189,9 +2311,12 @@ get_netdump_regs_x86(struct bt_info *bt, ulong *eip, ulong *esp)
 	char *sym;
 	ulong *up;
 	ulong ipintr_eip, ipintr_esp, ipintr_func;
-	ulong halt_eip, halt_esp;
+	ulong halt_eip, halt_esp, panic_eip, panic_esp;
 	int check_hardirq, check_softirq;
 	ulong stackbase, stacktop;
+	Elf32_Nhdr *note;
+	char *user_regs;
+	ulong ip, sp;
 
 	if (!is_task_active(bt->task)) {
 		machdep->get_stack_frame(bt, eip, esp);
@@ -2201,7 +2326,7 @@ get_netdump_regs_x86(struct bt_info *bt, ulong *eip, ulong *esp)
 	panic_task = tt->panic_task == bt->task ? TRUE : FALSE;
 
 	ipintr_eip = ipintr_esp = ipintr_func = panic = altered = 0;
-	halt_eip = halt_esp = 0;
+	halt_eip = halt_esp = panic_eip = panic_esp = 0;
 	check_hardirq = check_softirq = tt->flags & IRQSTACKS ? TRUE : FALSE;
 	search = ((bt->flags & BT_TEXT_SYMBOLS) && (tt->flags & TASK_INIT_DONE))
 		|| (machdep->flags & OMIT_FRAME_PTR);
@@ -2224,7 +2349,7 @@ retry:
 		} else if (STREQ(sym, "netconsole_netdump") || 
 		    STREQ(sym, "netpoll_start_netdump") ||
 		    STREQ(sym, "start_disk_dump") ||
-		    STREQ(sym, "crash_kexec") ||
+		    (STREQ(sym, "crash_kexec") && !KVMDUMP_DUMPFILE()) ||
 		    STREQ(sym, "disk_dump")) {
 			*eip = *up;
 			*esp = search ?
@@ -2238,6 +2363,8 @@ retry:
                         *esp = search ?
 			    bt->stackbase + ((char *)(up+1) - bt->stackbuf) :
 				*(up-1);
+			panic_eip = *eip;
+			panic_esp = *esp;
 			panic = TRUE;
                         continue;   /* keep looking for die */
                 }
@@ -2265,7 +2392,8 @@ next_sysrq:
                 		if (STREQ(sym, "sysrq_handle_crash")) 
 					goto next_sysrq; 
 			}
-                        return;
+			if (!panic)
+				return;
                 }
 
 		/* 
@@ -2288,6 +2416,14 @@ next_sysrq:
                         return;
                 }
 
+                if (STREQ(sym, "stop_this_cpu")) {
+                        *eip = *up;
+                        *esp = search ?
+                            bt->stackbase + ((char *)(up+1) - bt->stackbuf) :
+                                *(up-1);
+                        return;
+                }
+
                 if (STREQ(sym, "smp_call_function_interrupt")) {
 			if (ipintr_eip && IS_VMALLOC_ADDR(ipintr_func) &&
 		  	    IS_KERNEL_STATIC_TEXT(*(up - 2)))
@@ -2312,6 +2448,12 @@ next_sysrq:
                 }
 	}
 
+	if (panic) {
+		*eip = panic_eip;
+		*esp = panic_esp;
+		return;
+	}
+
 	if (ipintr_eip) {
         	*eip = ipintr_eip;
         	*esp = ipintr_esp;
@@ -2324,9 +2466,6 @@ next_sysrq:
 		return;
 	}
 
-	if (panic)
-		return;
-
 	bt->flags &= ~(BT_HARDIRQ|BT_SOFTIRQ);
 
 	if (check_hardirq &&
@@ -2351,6 +2490,20 @@ next_sysrq:
                 goto retry;
         }
 
+	if (ELF_NOTES_VALID() && DISKDUMP_DUMPFILE() &&
+	    (note = (Elf32_Nhdr *)
+	     diskdump_get_prstatus_percpu(bt->tc->processor))) {
+		user_regs = get_regs_from_note((char *)note, &ip, &sp);
+		if (is_kernel_text(ip) &&
+		    (((sp >= GET_STACKBASE(bt->task)) &&
+		      (sp < GET_STACKTOP(bt->task))) ||
+		    in_alternate_stack(bt->tc->processor, sp))) {
+			*eip = ip;
+			*esp = sp;
+			return;
+		}
+	}
+
 	if (CRASHDEBUG(1))
 		error(INFO, 
     "get_netdump_regs_x86: cannot find anything useful (task: %lx)\n", bt->task);
@@ -2367,6 +2520,9 @@ next_sysrq:
 		    "starting backtrace locations of the active (non-crashing) "
 		    "xen tasks\n    cannot be determined: try -t or -T options\n");
  
+	if (KVMDUMP_DUMPFILE())
+		bt->flags &= ~(ulonglong)BT_DUMPFILE_SEARCH;
+
 	machdep->get_stack_frame(bt, eip, esp);
 }
 
@@ -2705,10 +2861,41 @@ int get_netdump_arch(void)
 	return e_machine;
 }
 
+int 
+exist_regs_in_elf_notes(struct task_context *tc)
+{
+	if ((tc->task == tt->panic_task) ||
+	    (is_task_active(tc->task) && (nd->num_prstatus_notes > 1) &&
+	     (tc->processor < nd->num_prstatus_notes)))
+		return TRUE;
+	else
+		return FALSE;
+}
+
 void * 
 get_regs_from_elf_notes(struct task_context *tc)
 {
-	switch(get_netdump_arch())
+	int e_machine = get_netdump_arch();
+
+	switch (e_machine)
+	{
+	case EM_386:
+	case EM_PPC64:
+	case EM_X86_64:
+	case EM_ARM:
+		break;
+	default:
+		error(FATAL,
+		      "support for ELF machine type %d not available\n",
+		      e_machine);
+	}
+
+	if (!exist_regs_in_elf_notes(tc))
+		error(FATAL, "cannot determine register set "
+		      "for active task: %lx comm: \"%s\"\n",
+		      tc->task, tc->comm);
+
+	switch(e_machine)
 	{
 	case EM_386:
 		return get_x86_regs_from_elf_notes(tc);
@@ -2718,10 +2905,6 @@ get_regs_from_elf_notes(struct task_context *tc)
 		return get_x86_64_regs_from_elf_notes(tc);
 	case EM_ARM:
 		return get_arm_regs_from_elf_notes(tc);
-	default:
-		error(FATAL,
-		    "support for ELF machine type %d not available\n",
-			get_netdump_arch());
 	}
 
 	return NULL;
@@ -2739,31 +2922,25 @@ get_x86_regs_from_elf_notes(struct task_context *tc)
 	len = 0;
 	pt_regs = NULL;
 
-	if ((tc->task == tt->panic_task) || 
-	    (is_task_active(tc->task) && (nd->num_prstatus_notes > 1))) {
-		if (nd->num_prstatus_notes > 1)
-			note = (void *)
-				nd->nt_prstatus_percpu[tc->processor];
-		else
-			note = (void *)nd->nt_prstatus;
-		if (nd->elf32) {
-			note_32 = (Elf32_Nhdr *)note;
-			len = sizeof(Elf32_Nhdr);
-			len = roundup(len + note_32->n_namesz, 4);
-		} else if (nd->elf64) {
-			note_64 = (Elf64_Nhdr *)note;
-			len = sizeof(Elf64_Nhdr);
-			len = roundup(len + note_64->n_namesz, 4);
-		}
-		
-		pt_regs = (void *)((char *)note + len + 
-			MEMBER_OFFSET("elf_prstatus", "pr_reg"));
-		/* NEED TO BE FIXED: Hack to get the proper alignment */
-		pt_regs +=4;
-	} else
-		error(FATAL, 
-		    "cannot determine register set for task \"%s\"\n",
-			tc->comm);
+	if (nd->num_prstatus_notes > 1)
+		note = (void *)nd->nt_prstatus_percpu[tc->processor];
+	else
+		note = (void *)nd->nt_prstatus;
+	if (nd->elf32) {
+		note_32 = (Elf32_Nhdr *)note;
+		len = sizeof(Elf32_Nhdr);
+		len = roundup(len + note_32->n_namesz, 4);
+	} else if (nd->elf64) {
+		note_64 = (Elf64_Nhdr *)note;
+		len = sizeof(Elf64_Nhdr);
+		len = roundup(len + note_64->n_namesz, 4);
+	}
+
+	pt_regs = (void *)((char *)note + len +
+			   MEMBER_OFFSET("elf_prstatus", "pr_reg"));
+	/* NEED TO BE FIXED: Hack to get the proper alignment */
+	pt_regs +=4;
+
 	return pt_regs;
 
 }
@@ -2777,22 +2954,16 @@ get_x86_64_regs_from_elf_notes(struct task_context *tc)
 
 	pt_regs = NULL;
 
-	if ((tc->task == tt->panic_task) || 
-	    (is_task_active(tc->task) && (nd->num_prstatus_notes > 1))) {
-		if (nd->num_prstatus_notes > 1)
-			note = (Elf64_Nhdr *)
-				nd->nt_prstatus_percpu[tc->processor];
-		else
-			note = (Elf64_Nhdr *)nd->nt_prstatus;
+	if (nd->num_prstatus_notes > 1)
+		note = (Elf64_Nhdr *)nd->nt_prstatus_percpu[tc->processor];
+	else
+		note = (Elf64_Nhdr *)nd->nt_prstatus;
+
+	len = sizeof(Elf64_Nhdr);
+	len = roundup(len + note->n_namesz, 4);
+	pt_regs = (void *)((char *)note + len +
+			   MEMBER_OFFSET("elf_prstatus", "pr_reg"));
 
-		len = sizeof(Elf64_Nhdr);
-		len = roundup(len + note->n_namesz, 4);
-		pt_regs = (void *)((char *)note + len + 
-			MEMBER_OFFSET("elf_prstatus", "pr_reg"));
-	} else
-		error(FATAL, 
-		    "cannot determine register set for task \"%s\"\n",
-			tc->comm);
 	return pt_regs;
 }
 
@@ -2806,32 +2977,21 @@ get_ppc64_regs_from_elf_notes(struct task_context *tc)
 
 	pt_regs = NULL;
 
-	if ((tc->task == tt->panic_task) ||
-	    (is_task_active(tc->task) && (nd->num_prstatus_notes > 1))) {
-		/*	
-		 * Registers are always saved during the dump process for the 
-		 * panic task.  Kdump also captures registers for all CPUs if
-		 * they responded to an IPI.
-		 */
-		if (nd->num_prstatus_notes > 1) {
-			if (tc->processor >= nd->num_prstatus_notes)
-				error(FATAL, "cannot determine NT_PRSTATUS ELF note "
-				    "for %s task: %lx\n", (tc->task == tt->panic_task) ?
-				    "panic" : "active", tc->task);	
-			note = (Elf64_Nhdr *)
-				nd->nt_prstatus_percpu[tc->processor];
-		} else
-			note = (Elf64_Nhdr *)nd->nt_prstatus;
-
-		len = sizeof(Elf64_Nhdr);
-		len = roundup(len + note->n_namesz, 4);
-		pt_regs = (void *)((char *)note + len + 
-			MEMBER_OFFSET("elf_prstatus", "pr_reg"));
+	/*
+	 * Registers are always saved during the dump process for the
+	 * panic task.  Kdump also captures registers for all CPUs if
+	 * they responded to an IPI.
+	 */
+	if (nd->num_prstatus_notes > 1) {
+		note = (Elf64_Nhdr *)nd->nt_prstatus_percpu[tc->processor];
 	} else
-		error(FATAL, 
-		    "cannot determine register set for task \"%s\"\n",
-			tc->comm);
-	
+		note = (Elf64_Nhdr *)nd->nt_prstatus;
+
+	len = sizeof(Elf64_Nhdr);
+	len = roundup(len + note->n_namesz, 4);
+	pt_regs = (void *)((char *)note + len +
+			   MEMBER_OFFSET("elf_prstatus", "pr_reg"));
+
 	return pt_regs;
 }
 
@@ -2874,28 +3034,317 @@ get_arm_regs_from_elf_notes(struct task_context *tc)
 	len = 0;
 	pt_regs = NULL;
 
-	if ((tc->task == tt->panic_task) ||
-	    (is_task_active(tc->task) && (nd->num_prstatus_notes > 1))) {
-		if (nd->num_prstatus_notes > 1)
-			note = (void *)
-				nd->nt_prstatus_percpu[tc->processor];
-		else
-			note = (void *)nd->nt_prstatus;
-		if (nd->elf32) {
-			note_32 = (Elf32_Nhdr *)note;
-			len = sizeof(Elf32_Nhdr);
-			len = roundup(len + note_32->n_namesz, 4);
-		} else if (nd->elf64) {
-			note_64 = (Elf64_Nhdr *)note;
-			len = sizeof(Elf64_Nhdr);
-			len = roundup(len + note_64->n_namesz, 4);
-		}
+	if (nd->num_prstatus_notes > 1)
+		note = (void *)nd->nt_prstatus_percpu[tc->processor];
+	else
+		note = (void *)nd->nt_prstatus;
+	if (nd->elf32) {
+		note_32 = (Elf32_Nhdr *)note;
+		len = sizeof(Elf32_Nhdr);
+		len = roundup(len + note_32->n_namesz, 4);
+	} else if (nd->elf64) {
+		note_64 = (Elf64_Nhdr *)note;
+		len = sizeof(Elf64_Nhdr);
+		len = roundup(len + note_64->n_namesz, 4);
+	}
+
+	pt_regs = (void *)((char *)note + len +
+			   MEMBER_OFFSET("elf_prstatus", "pr_reg"));
 
-		pt_regs = (void *)((char *)note + len +
-			MEMBER_OFFSET("elf_prstatus", "pr_reg"));
-	} else
-		error(FATAL,
-		    "cannot determine arm register set for task \"%s\"\n",
-			tc->comm);
 	return pt_regs;
 }
+
+/*
+ *  Read from /proc/kcore.
+ */
+int
+read_proc_kcore(int fd, void *bufptr, int cnt, ulong addr, physaddr_t paddr) 
+{
+	int i; 
+	size_t readcnt;
+	ulong kvaddr;
+	Elf32_Phdr *lp32;
+	Elf64_Phdr *lp64;
+	off_t offset;
+
+	if (!machdep->verify_paddr(paddr)) {
+		if (CRASHDEBUG(1))
+			error(INFO, "verify_paddr(%lx) failed\n", paddr);
+		return READ_ERROR;
+	}
+
+	/*
+	 *  Turn the physical address into a unity-mapped kernel 
+	 *  virtual address, which should work for 64-bit architectures,
+	 *  and for lowmem access for 32-bit architectures.
+	 */
+	offset = UNINITIALIZED;
+	kvaddr = (ulong)paddr | machdep->kvbase;
+	readcnt = cnt;
+
+	switch (pkd->flags & (KCORE_ELF32|KCORE_ELF64)) 
+	{
+	case KCORE_ELF32:
+		for (i = 0; i < pkd->segments; i++) {
+			lp32 = pkd->load32 + i;
+			if ((kvaddr >= lp32->p_vaddr) &&
+			    (kvaddr < (lp32->p_vaddr + lp32->p_memsz))) {
+				offset = (off_t)(kvaddr - lp32->p_vaddr) + 
+					(off_t)lp32->p_offset;
+				break;
+			}
+		}
+		/*
+		 *  If it's not accessible via unity-mapping, check whether
+		 *  it's a request for a vmalloc address that can be found 
+                 *  in the header.
+		 */
+		if (pc->curcmd_flags & MEMTYPE_KVADDR)
+			pc->curcmd_flags &= ~MEMTYPE_KVADDR;
+		else
+			break;
+
+		for (i = 0; i < pkd->segments; i++) {
+			lp32 = pkd->load32 + i;
+			if ((addr >= lp32->p_vaddr) &&
+			    (addr < (lp32->p_vaddr + lp32->p_memsz))) {
+				offset = (off_t)(addr - lp32->p_vaddr) + 
+					(off_t)lp32->p_offset;
+				break;
+			}
+		}
+
+		break;
+
+	case KCORE_ELF64:
+		for (i = 0; i < pkd->segments; i++) {
+			lp64 = pkd->load64 + i;
+			if ((kvaddr >= lp64->p_vaddr) &&
+			    (kvaddr < (lp64->p_vaddr + lp64->p_memsz))) {
+				offset = (off_t)(kvaddr - lp64->p_vaddr) + 
+					(off_t)lp64->p_offset;
+				break;
+			}
+		}
+
+		break;
+	}
+
+	if (offset == UNINITIALIZED)
+		return SEEK_ERROR;
+
+        if (lseek(fd, offset, SEEK_SET) != offset)
+		perror("lseek");
+
+	if (read(fd, bufptr, readcnt) != readcnt)
+		return READ_ERROR;
+
+	return cnt;
+}
+
+/*
+ *  place holder -- cannot write to /proc/kcore
+ */
+int
+write_proc_kcore(int fd, void *bufptr, int cnt, ulong addr, physaddr_t paddr)
+{
+	error(FATAL, "cannot write to /proc/kcore\n");
+	return FALSE;
+}
+
+int
+is_proc_kcore(char *file, ulong source_query)
+{
+	if (STREQ(file, "/proc/kcore") || same_file(file, "/proc/kcore")) {
+		if (!is_netdump(file, source_query))
+			error(FATAL, 
+			    "cannot translate the ELF header of /proc/kcore\n");
+		pkd->flags |= KCORE_LOCAL;
+		return TRUE;
+	} else
+		return FALSE;
+}
+
+int
+proc_kcore_init(FILE *fp)
+{
+	if (BITS32())
+		return proc_kcore_init_32(fp);
+	else 
+		return proc_kcore_init_64(fp);
+}
+
+static int
+proc_kcore_init_32(FILE *fp)
+{
+	Elf32_Ehdr *elf32;
+	Elf32_Phdr *load32;
+	char eheader[MAX_KCORE_ELF_HEADER_SIZE];
+	char buf[BUFSIZE];
+	size_t size;
+
+	size = MAX_KCORE_ELF_HEADER_SIZE;
+
+	if (read(pc->mfd, eheader, size) != size) {
+		sprintf(buf, "/proc/kcore: read");
+		perror(buf);
+		goto bailout;
+	}
+
+	if (lseek(pc->mfd, 0, SEEK_SET) != 0) {
+		sprintf(buf, "/proc/kcore: lseek");
+		perror(buf);
+		goto bailout;
+	}
+
+	elf32 = (Elf32_Ehdr *)&eheader[0];
+	load32 = (Elf32_Phdr *)&eheader[sizeof(Elf32_Ehdr)+sizeof(Elf32_Phdr)];
+
+	pkd->segments = elf32->e_phnum - 1;
+
+	size = (ulong)(load32+(elf32->e_phnum)) - (ulong)elf32;
+	if ((pkd->elf_header = (char *)malloc(size)) == NULL) {
+		error(INFO, "/proc/kcore: cannot malloc ELF header buffer\n");
+		clean_exit(1);
+	}
+
+	BCOPY(&eheader[0], &pkd->elf_header[0], size);	
+	pkd->elf32 = (Elf32_Ehdr *)pkd->elf_header;
+	pkd->load32 = (Elf32_Phdr *)
+		&pkd->elf_header[sizeof(Elf32_Ehdr)+sizeof(Elf32_Phdr)];
+	pkd->flags |= KCORE_ELF32;
+	
+	if (CRASHDEBUG(1))
+		kcore_memory_dump(fp);
+
+	return TRUE;
+bailout:
+	return FALSE;
+}
+
+static int
+proc_kcore_init_64(FILE *fp)
+{
+	Elf64_Ehdr *elf64;
+	Elf64_Phdr *load64;
+	char eheader[MAX_KCORE_ELF_HEADER_SIZE];
+	char buf[BUFSIZE];
+	size_t size;
+
+	size = MAX_KCORE_ELF_HEADER_SIZE;
+
+	if (read(pc->mfd, eheader, size) != size) {
+		sprintf(buf, "/proc/kcore: read");
+		perror(buf);
+		goto bailout;
+	}
+
+	if (lseek(pc->mfd, 0, SEEK_SET) != 0) {
+		sprintf(buf, "/proc/kcore: lseek");
+		perror(buf);
+		goto bailout;
+	}
+
+	elf64 = (Elf64_Ehdr *)&eheader[0];
+	load64 = (Elf64_Phdr *)&eheader[sizeof(Elf64_Ehdr)+sizeof(Elf64_Phdr)];
+
+	pkd->segments = elf64->e_phnum - 1;
+
+	size = (ulong)(load64+(elf64->e_phnum)) - (ulong)elf64;
+	if ((pkd->elf_header = (char *)malloc(size)) == NULL) {
+		error(INFO, "/proc/kcore: cannot malloc ELF header buffer\n");
+		clean_exit(1);
+	}
+
+	BCOPY(&eheader[0], &pkd->elf_header[0], size);	
+	pkd->elf64 = (Elf64_Ehdr *)pkd->elf_header;
+	pkd->load64 = (Elf64_Phdr *)
+		&pkd->elf_header[sizeof(Elf64_Ehdr)+sizeof(Elf64_Phdr)];
+	pkd->flags |= KCORE_ELF64;
+	
+	if (CRASHDEBUG(1))
+		kcore_memory_dump(fp);
+
+	return TRUE;
+bailout:
+	return FALSE;
+}
+
+int
+kcore_memory_dump(FILE *ofp)
+{
+	int i, others;
+	Elf32_Phdr *lp32;
+	Elf64_Phdr *lp64;
+
+	if (!(pkd->flags & KCORE_LOCAL))
+		return FALSE;
+
+	fprintf(ofp, "proc_kcore_data:\n");
+	fprintf(ofp, "       flags: %lx (", nd->flags);
+	others = 0;
+	if (pkd->flags & KCORE_LOCAL)
+		fprintf(ofp, "%sKCORE_LOCAL", others++ ? "|" : "");
+	if (pkd->flags & KCORE_ELF32)
+		fprintf(ofp, "%sKCORE_ELF32", others++ ? "|" : "");
+	if (pkd->flags & KCORE_ELF64)
+		fprintf(ofp, "%sKCORE_ELF64", others++ ? "|" : "");
+	fprintf(ofp, ")\n");
+	fprintf(ofp, "    segments: %d\n",
+		pkd->segments);
+	fprintf(ofp, "  elf_header: %lx\n", (ulong)pkd->elf_header);
+	fprintf(ofp, "       elf64: %lx\n", (ulong)pkd->elf64);
+	fprintf(ofp, "      load64: %lx\n", (ulong)pkd->load64);
+	fprintf(ofp, "       elf32: %lx\n", (ulong)pkd->elf32);
+	fprintf(ofp, "      load32: %lx\n\n", (ulong)pkd->load32);
+
+	for (i = 0; i < pkd->segments; i++) {
+		if (pkd->flags & KCORE_ELF32)
+			break;
+
+		lp64 = pkd->load64 + i;
+
+		fprintf(ofp, "  Elf64_Phdr:\n");
+		fprintf(ofp, "        p_type: %x\n", lp64->p_type);
+		fprintf(ofp, "       p_flags: %x\n", lp64->p_flags);
+		fprintf(ofp, "      p_offset: %llx\n", (ulonglong)lp64->p_offset);
+		fprintf(ofp, "       p_vaddr: %llx\n", (ulonglong)lp64->p_vaddr);
+		fprintf(ofp, "       p_paddr: %llx\n", (ulonglong)lp64->p_paddr);
+		fprintf(ofp, "      p_filesz: %llx\n", (ulonglong)lp64->p_filesz);
+		fprintf(ofp, "       p_memsz: %llx\n", (ulonglong)lp64->p_memsz);
+		fprintf(ofp, "       p_align: %lld\n", (ulonglong)lp64->p_align);
+		fprintf(ofp, "\n");
+	}
+
+	for (i = 0; i < pkd->segments; i++) {
+		if (pkd->flags & KCORE_ELF64)
+			break;
+
+		lp32 = pkd->load32 + i;
+
+		fprintf(ofp, "  Elf32_Phdr:\n");
+		fprintf(ofp, "        p_type: %x\n", lp32->p_type);
+		fprintf(ofp, "       p_flags: %x\n", lp32->p_flags);
+		fprintf(ofp, "      p_offset: %x\n", lp32->p_offset);
+		fprintf(ofp, "       p_vaddr: %x\n", lp32->p_vaddr);
+		fprintf(ofp, "       p_paddr: %x\n", lp32->p_paddr);
+		fprintf(ofp, "      p_filesz: %x\n", lp32->p_filesz);
+		fprintf(ofp, "       p_memsz: %x\n", lp32->p_memsz);
+		fprintf(ofp, "       p_align: %d\n", lp32->p_align);
+		fprintf(ofp, "\n");
+	}
+
+	return TRUE;
+}
+
+static void
+kdump_get_osrelease(void)
+{
+	char *string;
+
+	if ((string = vmcoreinfo_read_string("OSRELEASE"))) {
+		fprintf(fp, "%s\n", string);
+		free(string);
+	} else 
+		pc->flags2 &= ~GET_OSRELEASE;
+}
