@@ -1,7 +1,7 @@
 /* ppc64.c -- core analysis suite
  *
- * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010 David Anderson
- * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010 Red Hat, Inc. All rights reserved.
+ * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 David Anderson
+ * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Red Hat, Inc. All rights reserved.
  * Copyright (C) 2004, 2006 Haren Myneni, IBM Corporation
  *
  * This program is free software; you can redistribute it and/or modify
@@ -52,6 +52,7 @@ static void parse_cmdline_args(void);
 static void ppc64_paca_init(void);
 static void ppc64_clear_machdep_cache(void);
 static void ppc64_vmemmap_init(void);
+static int ppc64_get_kvaddr_ranges(struct vaddr_range *);
 
 struct machine_specific ppc64_machine_specific = { { 0 }, 0, 0 };
 
@@ -137,6 +138,7 @@ ppc64_init(int when)
 		machdep->get_smp_cpus = ppc64_get_smp_cpus;
 		machdep->line_number_hooks = ppc64_line_number_hooks;
 		machdep->value_to_symbol = generic_machdep_value_to_symbol;
+		machdep->get_kvaddr_ranges = ppc64_get_kvaddr_ranges;
 		machdep->init_kernel_pgd = NULL;
 		if (symbol_exists("vmemmap_populate")) {
 			machdep->flags |= VMEMMAP;
@@ -373,6 +375,7 @@ ppc64_dump_machdep_table(ulong arg)
         fprintf(fp, "          is_kvaddr: generic_is_kvaddr()\n");
         fprintf(fp, "          is_uvaddr: generic_is_uvaddr()\n");
         fprintf(fp, "       verify_paddr: generic_verify_paddr()\n");
+        fprintf(fp, "  get_kvaddr_ranges: ppc64_get_kvaddr_ranges()\n");
 	fprintf(fp, " xendump_p2m_create: NULL\n");
 	fprintf(fp, "xen_kdump_p2m_create: NULL\n");
         fprintf(fp, "  line_number_hooks: ppc64_line_number_hooks\n");
@@ -1561,6 +1564,7 @@ ppc64_print_stack_entry(int frame,
 		      ulong lr, 	
 		      struct bt_info *bt)
 {
+	struct load_module *lm;
 	char *lrname = NULL;
 	if (BT_REFERENCE_CHECK(bt)) {
 		switch (bt->ref->cmdflags & (BT_REF_SYMBOL|BT_REF_HEXVAL))
@@ -1579,6 +1583,8 @@ ppc64_print_stack_entry(int frame,
 		fprintf(fp, "%s#%d [%lx] %s at %lx",
 			frame < 10 ? " " : "", frame,
 			req->sp, req->name, req->pc);
+		if (module_symbol(req->pc, NULL, &lm, NULL, 0))
+			fprintf(fp, " [%s]", lm->mod_name);
 	
 		if (req->ra) {
 			/*
@@ -2324,6 +2330,7 @@ ppc64_cmd_mach(void)
 static void
 ppc64_display_machine_stats(void)
 {
+	int c;
         struct new_utsname *uts;
         char buf[BUFSIZE];
         ulong mhz;
@@ -2344,6 +2351,27 @@ ppc64_display_machine_stats(void)
         fprintf(fp, "KERNEL VIRTUAL BASE: %lx\n", machdep->kvbase);
         fprintf(fp, "KERNEL VMALLOC BASE: %lx\n", vt->vmalloc_start);
         fprintf(fp, "  KERNEL STACK SIZE: %ld\n", STACKSIZE());
+
+	if (tt->flags & IRQSTACKS) {
+		fprintf(fp, "HARD IRQ STACK SIZE: %ld\n", STACKSIZE());
+		fprintf(fp, "    HARD IRQ STACKS:\n");
+	
+		for (c = 0; c < kt->cpus; c++) {
+			if (!tt->hardirq_ctx[c])
+				break;
+			sprintf(buf, "CPU %d", c);
+			fprintf(fp, "%19s: %lx\n", buf, tt->hardirq_ctx[c]);
+		}
+
+		fprintf(fp, "SOFT IRQ STACK SIZE: %ld\n", STACKSIZE());
+		fprintf(fp, "    SOFT IRQ STACKS:\n");
+		for (c = 0; c < kt->cpus; c++) {
+			if (!tt->softirq_ctx)
+				break;
+			sprintf(buf, "CPU %d", c);
+			fprintf(fp, "%19s: %lx\n", buf, tt->softirq_ctx[c]);
+		}
+	}
 }
 
 static const char *hook_files[] = {
@@ -2652,4 +2680,35 @@ ppc64_clear_machdep_cache(void)
 	if (machdep->machspec->last_level4_read != vt->kernel_pgd[0])
         	machdep->machspec->last_level4_read = 0;
 }
+
+static int 
+ppc64_get_kvaddr_ranges(struct vaddr_range *vrp)
+{
+	int cnt;
+	physaddr_t phys1, phys2;
+	ulong pp1, pp2;
+
+	cnt = 0;
+
+	vrp[cnt].type = KVADDR_UNITY_MAP;
+	vrp[cnt].start = machdep->kvbase;
+	vrp[cnt++].end = vt->high_memory;
+
+	vrp[cnt].type = KVADDR_VMALLOC;
+	vrp[cnt].start = first_vmalloc_address();
+	vrp[cnt++].end = last_vmalloc_address();
+
+	if (machdep->flags & VMEMMAP) {
+ 		phys1 = (physaddr_t)(0);
+		phys2 = (physaddr_t)VTOP((vt->high_memory - PAGESIZE()));
+		if (phys_to_page(phys1, &pp1) && 
+	    	    phys_to_page(phys2, &pp2)) {
+			vrp[cnt].type = KVADDR_VMEMMAP;
+			vrp[cnt].start = pp1;
+			vrp[cnt++].end = pp2;
+		}
+	}
+
+	return cnt;
+}
 #endif /* PPC64 */ 
